from typing import Optional, Tuple, Dict, Any, TYPE_CHECKING

from PySide6.QtGui import Qt
from PySide6.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QLabel, QPushButton, QDialogButtonBox, QLineEdit

import pycparser

from angr.sim_variable import SimVariable
from angr.analyses.decompiler.structured_codegen.c import CVariable, CConstruct
import angr


if TYPE_CHECKING:
    from angr.sim_type import SimType
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.views.code_view import CodeView


class TypeBox(QLineEdit):
    """
    Implements a line edit widget for inputting types.
    """
    def __init__(self, textchanged_callback, predefined_types=None, parent=None):
        super().__init__(parent)

        self._cvariable: Optional[CVariable] = None
        self._predefined_types = predefined_types


        self.textChanged.connect(textchanged_callback)

    def set_type(self, type_: 'SimType', cvariable: CVariable=None):
        self._cvariable = cvariable
        if cvariable is not None and isinstance(cvariable.unified_variable, SimVariable):
            type_str = type_.c_repr(name=cvariable.unified_variable.name)
        else:
            type_str = type_.c_repr(name="foobar")
        self.setText(type_str)

    @property
    def type_str(self):
        text = self.text()
        valid, _ = self._is_valid_type_str(text)
        if valid:
            return text.strip()
        return None

    @property
    def sim_type(self):
        text = self.text()
        valid, parsed_type = self._is_valid_type_str(text)
        if valid:
            return parsed_type
        return None

    def _is_valid_type_str(self, type_str: str) -> Tuple[bool,Optional['SimType']]:
        """
        We accept two forms of type strings. The user can either specify a full variable declaration, like "char var",
        or only specify a type string, like "char". This method first attempts to parse the string as a variable
        declaration, and when it fails, it will parse the string as a pure type string.
        """

        # parse as a variable declaration
        declaration = type_str
        if not declaration.endswith(";"):
            declaration += ";"
        try:
            defns = angr.sim_type.parse_defns(declaration, predefined_types=self._predefined_types)
        except pycparser.c_parser.ParseError:
            defns = { }
        if len(defns) == 1:
            return True, next(iter(defns.values()))

        try:
            parsed_type = angr.sim_type.parse_type(type_str, predefined_types=self._predefined_types)
            return True, parsed_type
        except pycparser.c_parser.ParseError:
            return False, None


class RetypeNode(QDialog):
    """
    A dialog for retyping nodes.
    """
    def __init__(self, instance: 'Instance', code_view: Optional['CodeView']=None, node: Optional[CConstruct]=None,
                 node_type=None, variable=None, parent=None):
        super().__init__(parent)

        # initialization
        self._instance = instance
        self._code_view = code_view
        self._node = node
        self._node_type = node_type
        self._variable = variable

        self.new_type = None

        self._type_box: TypeBox = None
        self._status_label = None
        self._ok_button: QPushButton = None

        self.setWindowTitle('Specify a type')
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        self.main_layout = QVBoxLayout()

        self._init_widgets()

        self.setLayout(self.main_layout)

    #
    # Private methods
    #

    def _get_predefined_types(self) -> Dict[Any,'SimType']:
        # global types
        r = dict(self._instance.kb.types)
        # local types
        if not self._code_view.function.am_none:
            func_addr = self._code_view.function.addr
            if (func_addr, 'pseudocode') in self._instance.kb.structured_code:
                pseudocode_cache = self._instance.kb.structured_code[(func_addr, 'pseudocode')]
                r.update(pseudocode_cache.local_types)
        return r

    def _init_widgets(self):

        # type label

        type_label = QLabel(self)
        type_label.setText('New type')

        type_box = TypeBox(self._on_type_changed, predefined_types=self._get_predefined_types(), parent=self)
        if self._node is not None:
            if isinstance(self._node, CVariable) and self._node.unified_variable:
                if self._node_type is not None:
                    type_box.set_type(self._node_type, cvariable=self._node)

            type_box.selectAll()
        self._type_box = type_box

        label_layout = QHBoxLayout()
        label_layout.addWidget(type_label)
        label_layout.addWidget(type_box)
        self.main_layout.addLayout(label_layout)

        # status label
        status_label = QLabel(self)
        self.main_layout.addWidget(status_label)
        self._status_label = status_label

        # buttons
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self._on_cancel_clicked)
        self._ok_button = buttons.button(QDialogButtonBox.Ok)
        self._ok_button.setEnabled(False)
        self.main_layout.addWidget(buttons)

    #
    # Event handlers
    #

    def _on_type_changed(self, new_text):  # pylint:disable=unused-argument
        if self._type_box is None:
            # initialization is not done yet
            return

        if self._type_box.type_str is None:
            # the variable name is invalid
            self._status_label.setText('Invalid')
            self._status_label.setProperty('class', 'status_invalid')
            self._ok_button.setEnabled(False)
        else:
            self._status_label.setText('Valid')
            self._status_label.setProperty('class', 'status_valid')
            self._ok_button.setEnabled(True)

        self._status_label.style().unpolish(self._status_label)
        self._status_label.style().polish(self._status_label)

    def _on_ok_clicked(self):
        self.new_type = self._type_box.sim_type
        self.close()

    def _on_cancel_clicked(self):
        self.new_type = None
        self.close()



from collections import defaultdict
from typing import Dict, List, Optional, Sequence
import logging
import functools

import PySide6QtAds as QtAds

from .views.view import BaseView


_l = logging.getLogger(__name__)


class ViewManager:
    """
    Manages views.
    """

    DOCKING_POSITIONS = {
        'center': QtAds.CenterDockWidgetArea,
        'left': QtAds.LeftDockWidgetArea,
        'right': QtAds.RightDockWidgetArea,
        'top': QtAds.TopDockWidgetArea,
        'bottom': QtAds.BottomDockWidgetArea,
    }

    def __init__(self, workspace):
        self.workspace = workspace
        self.views: List[BaseView] = [ ]
        self.docks = [ ]
        self.view_to_dock = { }
        self.dock_to_view = { }
        self.views_by_category: Dict[str,List[BaseView]] = defaultdict(list)

    @property
    def main_window(self):
        return self.workspace._main_window

    def main_window_initialized(self):
        """
        Invoked by the main window after it has finished initialization. Views can override
        BaseView.mainWindowInitializedEvent() to support delayed initialization or loading.
        """
        for view in self.views:
            view.mainWindowInitializedEvent()

    def _update_view_index_in_category(self, view: BaseView):
        """
        Set lowest available index value in category for a view not yet added.
        """
        existing_views = self.views_by_category[view.category]
        if view in existing_views:
            return
        existing_ids = {view.index for view in existing_views}
        max_id = max(existing_ids) if existing_ids else 0
        candidates = set(range(1, max_id + 2)) - existing_ids
        view.index = min(candidates)

    def add_view(self, view: BaseView):
        """
        Add a view to this workspace.

        :param view:            The view to add.
        :return:                None
        """
        self._update_view_index_in_category(view)
        self.views_by_category[view.category].append(view)

        dw = QtAds.CDockWidget(view.caption)
        dw.setFeature(QtAds.CDockWidget.DockWidgetDeleteOnClose, True)
        dw.closed.connect(functools.partial(self._on_dock_widget_closed, dw))
        dw.setWidget(view)

        area = self.DOCKING_POSITIONS.get(view.default_docking_position, QtAds.RightDockWidgetArea)
        self.main_window.central_widget.addDockWidgetTab(area, dw)

        self.views.append(view)
        self.docks.append(dw)
        self.view_to_dock[view] = dw
        self.dock_to_view[dw] = view

    def _on_dock_widget_closed(self, dock):
        """
        Handle dock widget close event.
        """
        if dock not in self.docks:
            return

        self.docks.remove(dock)
        view = self.dock_to_view.pop(dock, None)
        if view:
            view.close()
            self.remove_view(view)

    def remove_view(self, view: BaseView):
        """
        Remove a view from this workspace

        :param view: The view to remove.
        """
        if view not in self.views:
            return

        self.views.remove(view)
        self.views_by_category[view.category].remove(view)
        dock = self.view_to_dock.pop(view, None)
        if dock:
            dock.closeDockWidget()

    def raise_view(self, view: BaseView):
        """
        Find the dock widget of a view, and then bring that dock widget to front.

        :param BaseView view: The view to raise.
        :return:              None
        """

        # find the dock widget by the view
        dock = self.view_to_dock.get(view, None)
        if dock is None:
            return

        if not dock.isTabbed():
            dock.show()
        dock.raise_()
        view.focusWidget()

    def get_center_views(self) -> Sequence[QtAds.CDockWidget]:
        """
        Get the right dockable views

        :return:    Right Dockable Views
        """

        docks = []
        for dock in self.docks:
            if dock.widget() is not None:
                if dock.widget().default_docking_position == 'center':
                    docks.append(dock)
        return docks

    def first_view_in_category(self, category: str) -> Optional['BaseView']:
        """
        Return the first view in a specific category.

        :param str category:    The category of the view.
        :return:                The view.
        """

        if self.views_by_category[category]:
            return self.views_by_category[category][0]
        return None

    def current_view_in_category(self, category: str) -> Optional['BaseView']:
        """
        Return the current in a specific category.

        :param str category:    The category of the view.
        :return:                The view.
        """
        current_tab_id = self.get_current_tab_id()
        if current_tab_id is None:
            return None

        current = self.get_center_views()[current_tab_id]
        view = self.dock_to_view[current]
        if category.capitalize() in view.caption and view.caption == current.windowTitle():
            return view
        return None

    def get_current_tab_id(self) -> Optional[int]:
        """
        Get Current Tab ID

        :return:    The current tab ID, or None if no current tab exists in the central view area.
        """

        center_dockable_views = self.get_center_views()
        for i in range(1,len(center_dockable_views)+1):
            if center_dockable_views[i-1].visibleRegion().isEmpty() is False:
                return i-1
        return None

    def next_tab(self):
        """
        Shift to the next tab

        :return:    None
        """

        center_dockable_views = self.get_center_views()
        current_tab_id = self.get_current_tab_id()
        if current_tab_id is None:
            return
        if (current_tab_id + 1) < len(center_dockable_views):
            center_dockable_views[current_tab_id + 1].raise_()
        else:
            # Start from 1 again to prevent Index Out Of Range error
            center_dockable_views[(current_tab_id + 1) % (len(center_dockable_views))].raise_()

    def previous_tab(self):
        """
        Shift to the next tab

        :return:    None
        """

        current_tab_id = self.get_current_tab_id()
        if current_tab_id is None:
            return

        center_dockable_views = self.get_center_views()
        center_dockable_views[current_tab_id - 1].raise_()

    @property
    def current_tab(self) -> Optional['BaseView']:
        current_tab_id = self.get_current_tab_id()
        if current_tab_id is None:
            return None
        return self.get_center_views()[current_tab_id].widget()

    def _handle_raise_view(self, view: BaseView):
        self.workspace.plugins.handle_raise_view(view)



from PySide6.QtGui import QColor
from PySide6.QtWidgets import QWidget, QHBoxLayout, QFrame, QLabel, QColorDialog

from ...data.object_container import ObjectContainer


class QColorOption(QWidget):
    def __init__(self, color: QColor, label: str, parent=None):
        super().__init__(parent=parent)

        self.color = ObjectContainer(color, "The current color")
        self.label = label

        self._init_widgets()

    def set_color(self, color):
        self.color.am_obj = color
        self.color.am_event()

    def mouseReleaseEvent(self, event): # pylint:disable=unused-argument
        dialog = QColorDialog()
        dialog.setCurrentColor(self.color.am_obj)
        dialog.exec()
        if dialog.result() == QColorDialog.Accepted:
            self.set_color(dialog.currentColor())

    def _init_widgets(self):
        layout = QHBoxLayout()
        frame = QFrame()
        frame.setFixedWidth(30)
        frame.setFixedHeight(15)

        def update_color(**kwargs): # pylint:disable=unused-argument
            r,g,b,a = self.color.getRgb()
            frame.setStyleSheet(f"background-color: rgba({r},{g},{b},{a});")
        update_color()
        self.color.am_subscribe(update_color)

        text = QLabel(self.label)

        layout.addWidget(frame)
        layout.addWidget(text)
        self.setLayout(layout)



import re
from PySide6.QtWidgets import QCheckBox, QHBoxLayout, QLineEdit, QVBoxLayout, QLabel
from PySide6.QtCore import QSize

from angr.knowledge_plugins import Function
from angr.knowledge_plugins.cfg.memory_data import MemoryData

from .view import BaseView
from ..widgets.qstring_table import QStringTable
from ..widgets.qfunction_combobox import QFunctionComboBox


class StringsView(BaseView):
    def __init__(self, instance, default_docking_position, *args, **kwargs):
        super(StringsView, self).__init__('strings', instance, default_docking_position, *args, **kwargs)

        self.base_caption = 'Strings'

        self._string_table = None  # type: QStringTable
        self._function_list = None  # type: QFunctionComboBox

        self._selected_function = None

        self._init_widgets()
        self.reload()

    def reload(self):
        if self.instance.kb is None:
            return
        self._function_list.functions = self.instance.kb.functions
        self._string_table.cfg = self.instance.cfg
        self._string_table.xrefs = self.instance.project.kb.xrefs
        self._string_table.function = self._selected_function

    def sizeHint(self):
        return QSize(400, 800)

    #
    # Event handlers
    #

    def _on_function_selected(self, function):

        if isinstance(function, str) and function == "all":
            # all functions
            self._selected_function = None

        elif isinstance(function, Function):
            self._selected_function = function

        self.reload()

    def _on_string_selected(self, s: MemoryData):
        """
        A string reference is selected.

        :param s:
        :return:
        """

        if len(self.instance.workspace.view_manager.views_by_category['disassembly']) == 1:
            disasm_view = self.instance.workspace.view_manager.first_view_in_category('disassembly')
        else:
            disasm_view = self.instance.workspace.view_manager.current_view_in_category('disassembly')
        if disasm_view is not None:
            disasm_view.jump_to(s.addr)
            disasm_view.select_label(s.addr)
            self.instance.workspace.view_manager.raise_view(disasm_view)

    def on_filter_change(self, **kwargs): #pylint: disable=unused-argument
        pattern = self._filter_string.text()
        regex = self._regex_checkbox.isChecked()
        if regex:
            try:
                pattern = re.compile(pattern)
            except re.error as _e:
                return
        self._string_table.filter_string = pattern

    #
    # Private methods
    #

    def _init_widgets(self):

        self._function_list = QFunctionComboBox(show_all_functions=True, selection_callback=self._on_function_selected,
                                                parent=self
                                                )

        self._filter_string = QLineEdit(self)
        self._regex_checkbox = QCheckBox("Regex", self)
        self._filter_string.textChanged.connect(self.on_filter_change)
        self._regex_checkbox.stateChanged.connect(self.on_filter_change)

        function_layout = QHBoxLayout()
        function_layout.addWidget(QLabel("Function:",self))
        function_layout.addWidget(self._function_list, 10)
        function_layout.addWidget(QLabel("Filter:",self))
        function_layout.addWidget(self._filter_string, 10)
        function_layout.addWidget(self._regex_checkbox)
        function_layout.setContentsMargins(3, 3, 3, 3)
        function_layout.setSpacing(4)

        self._string_table = QStringTable(self.instance, self, selection_callback=self._on_string_selected)

        layout = QVBoxLayout()
        layout.addLayout(function_layout)
        layout.addWidget(self._string_table)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)



from typing import TYPE_CHECKING

from PySide6.QtGui import QKeySequence, Qt

from .menu import Menu, MenuEntry, MenuSeparator

if TYPE_CHECKING:
    from angrmanagement.ui.widgets.qlog_widget import QLogWidget


class LogMenu(Menu):
    def __init__(self, log_widget: 'QLogWidget'):
        super().__init__("", parent=log_widget)

        self.entries.extend([
            MenuEntry('&Copy selected content', log_widget.copy_selected_messages,
                      shortcut=QKeySequence(Qt.CTRL | Qt.Key_C)),
            MenuEntry('Copy selected message', log_widget.copy_selected),
            MenuEntry('Copy all content', log_widget.copy_all_messages),
            MenuEntry('Copy all messages', log_widget.copy_all),
            MenuSeparator(),
            MenuEntry('C&lear log', log_widget.clear_log),
        ])




from ..utils.graph import to_supergraph

def edge_qualifies(data):
    return data['type'] not in ('call', 'return_from_call')


class FunctionGraph:

    def __init__(self, function, exception_edges=True):
        self.function = function
        self.exception_edges = exception_edges
        self.edges = None
        self._supergraph = None

    def clear_cache(self):
        self._supergraph = None
        self.edges = None

    @property
    def supergraph(self):
        if self._supergraph is not None:
            return self._supergraph

        self._supergraph = to_supergraph(self.function.transition_graph_ex(exception_edges=self.exception_edges))
        self.edges = [(str(from_.addr), str(to.addr)) for (from_, to, data) in self._supergraph.edges(data=True) if
                      edge_qualifies(data)
                      ]

        return self._supergraph



import asyncio
import math
import threading
import time
from time import sleep
from typing import Optional


from PySide6.QtCore import QEvent, QObject
from PySide6.QtWidgets import QDialog, QLabel, QLineEdit, QPushButton, QVBoxLayout

try:
    from slacrs import Slacrs
    from slacrs.model import HumanFatigue
except ImportError as ex:
    Slacrs = None  # type: Optional[type]
    HumanFatigue = None  # type: Optional[type]
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

from ..base_plugin import BasePlugin

#
# Plugin to capture the User Mouse Movements.
# User must input their name to use this plugin.
#


class LogFatiguePlugin(BasePlugin):
    def __init__(self, workspace):
        if not Slacrs:
            raise Exception(
                "Please install Slacrs to Initialize LogFatigue Plugin"
            )
        self._fatigue_flag = True
        super().__init__(workspace)
        self._fatigue = HumanFatigue()
        self._strokes = []
        self._main_window = workspace.view_manager.main_window
        self._main_window.setMouseTracking(True)
        self.EventFilterInstance = self.EventFilter(self._fatigue)
        self._main_window.installEventFilter(self.EventFilterInstance)
        self.modal = self.Modal(self)
        self.modal.show()

        self.t_log = threading.Thread(target=self._log_mouse, args=())
        self.t_log.setDaemon(True)
        self.t_log.start()

    def _log_mouse(self):
        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())

        connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
        if connector is None:
            # chess connector does not exist
            return None
        slacrs_instance = connector.slacrs_instance()
        if slacrs_instance is None:
            # slacrs does not exist. continue
            return None
        session = slacrs_instance.session()
        with session.no_autoflush:
            while self._fatigue_flag is True:
                sleep(2)
                if self._fatigue.user:
                    session.add(self._fatigue)
                    session.commit()
            session.close()

    class EventFilter(QObject):
        def __init__(self, fatigue):
            super().__init__()
            self._fatigue = fatigue
            self._strokes = []

        def eventFilter(self, obj, event):  # pylint: disable=unused-argument
            if event.type() == QEvent.HoverMove:
                x = event.pos().x()
                old_x = event.oldPos().x()
                y = event.pos().y()
                old_y = event.oldPos().y()

                self._fatigue.mouse_speed = int(
                    math.sqrt((x - old_x) ** 2 + (y - old_y) ** 2)
                )
            elif event.type() == QEvent.KeyPress:
                timestamp = time.time()
                i = 0
                for i in range(len(self._strokes)):
                    if timestamp - self._strokes[i] <= 10:
                        break
                self._strokes = self._strokes[i:]
                self._strokes.append(timestamp)
                self._fatigue.stroke = len(self._strokes)

            return False

    #
    # Creates Model so user can input name, if modal is closed without submitting name,
    # the LogFatigue plugin will be deactivated
    #

    class Modal(QDialog):
        def __init__(self, outerclass):
            super().__init__(outerclass._main_window)

            self.label = QLabel("Enter your Name")
            self.edit = QLineEdit("")
            self.button = QPushButton("Submit")
            self.outerclass = outerclass

            layout = QVBoxLayout()

            layout.addWidget(self.label)
            layout.addWidget(self.edit)
            layout.addWidget(self.button)

            self.setLayout(layout)

            self.button.clicked.connect(self.input)

        def input(self):
            self.outerclass._fatigue.user = self.edit.text()
            self.close()

        def closeEvent(self, event):
            if not self.outerclass._fatigue.user:
                self.outerclass._fatigue_flag = False
                self.outerclass.workspace.plugins.deactivate_plugin(self.outerclass)
            event.accept()

    def teardown(self):
        self._fatigue_flag = False
        self.t_log.join()



# pylint:disable=unused-argument
import os
import logging
from typing import List, Any, Optional

import PySide6
from PySide6.QtWidgets import QTableView, QAbstractItemView, QHeaderView
from PySide6.QtCore import QAbstractTableModel, Qt
from PySide6.QtGui import QIcon, QCursor, QGuiApplication, QClipboard, QKeySequence

from ...config import IMG_LOCATION
from ...logic.threads import gui_thread_schedule_async
from ...data.log import LogRecord
from ..menus.log_menu import LogMenu


class QLogIcons:
    """
    All icons that are used in the log table.
    """

    BENCHMARK = None
    WARNING = None
    ERROR = None

    @classmethod
    def benchmark(cls) -> QIcon:
        if cls.BENCHMARK is None:
            cls.BENCHMARK = QIcon(os.path.join(IMG_LOCATION, 'benchmark-icon.png'))
        return cls.BENCHMARK

    @classmethod
    def warning(cls) -> QIcon:
        if cls.WARNING is None:
            cls.WARNING = QIcon(os.path.join(IMG_LOCATION, 'warning-icon.png'))
        return cls.WARNING

    @classmethod
    def error(cls) -> QIcon:
        if cls.ERROR is None:
            cls.ERROR = QIcon(os.path.join(IMG_LOCATION, 'error-icon.png'))
        return cls.ERROR


class QLogTableModel(QAbstractTableModel):
    """
    Implements a table model for log items.
    """

    Headers = ["", "Timestamp", "Source", "Content"]
    COL_ICON = 0
    COL_TIMESTAMP = 1
    COL_SOURCE = 2
    COL_CONTENT = 3

    def __init__(self, log_widget: 'QLogWidget'=None):
        super().__init__()
        self._log_widget = log_widget
        self._log: List[LogRecord] = [ ]

    @property
    def log(self) -> List[LogRecord]:
        return self._log

    def rowCount(self, parent:PySide6.QtCore.QModelIndex=...) -> int:
        return len(self._log)

    def columnCount(self, parent:PySide6.QtCore.QModelIndex=...) -> int:
        return len(self.Headers)

    def headerData(self, section:int, orientation:PySide6.QtCore.Qt.Orientation, role:int=...) -> Any:
        if role != Qt.DisplayRole:
            return None
        if section < len(self.Headers):
            return self.Headers[section]
        return None

    def data(self, index:PySide6.QtCore.QModelIndex, role:int=...) -> Any:
        if not index.isValid():
            return None
        row = index.row()
        if row >= len(self.log):
            return None
        log = self.log[row]
        col = index.column()

        if role == Qt.DisplayRole:
            return self._get_column_text(log, col)
        elif role == Qt.DecorationRole and col == QLogTableModel.COL_ICON:
            return self._get_column_icon(log)

        return None

    @staticmethod
    def _get_column_text(log: LogRecord, col: int) -> Any:
        mapping = {
            QLogTableModel.COL_TIMESTAMP: lambda x: str(x.timestamp),
            QLogTableModel.COL_SOURCE: lambda x: str(x.source),
            QLogTableModel.COL_CONTENT: lambda x: str(x.content),
        }
        func = mapping.get(col)
        if func is None:
            return None
        return func(log)

    @staticmethod
    def _get_column_icon(log: LogRecord) -> Optional[QIcon]:
        mapping = {
            1: QLogIcons.benchmark(),
            logging.WARNING: QLogIcons.warning(),
            logging.ERROR: QLogIcons.error(),
            logging.CRITICAL: QLogIcons.error(),
        }
        return mapping.get(log.level, None)

    @staticmethod
    def level_to_text(loglevel: int) -> str:
        mapping = {
            1: "BENCHMARK",
            logging.DEBUG: "DEBUG",
            logging.INFO: "INFO",
            logging.WARNING: "WARNING",
            logging.ERROR: "ERROR",
            logging.CRITICAL: "CRITICAL",
        }
        return mapping.get(loglevel, "")


class QLogWidget(QTableView):
    """
    Log table. Displays log messages.
    """

    def __init__(self, log_view, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.log_view = log_view
        # True if we want the log panel to automatically scroll to bottom
        self._auto_scroll: bool = True
        self._context_menu = LogMenu(self).qmenu()

        hheader = self.horizontalHeader()
        vheader = self.verticalHeader()
        hheader.setVisible(True)
        hheader.setStretchLastSection(True)
        vheader.setVisible(False)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        vheader.setDefaultSectionSize(20)
        self.setShowGrid(False)

        self.model: QLogTableModel = QLogTableModel(self)

        self.setModel(self.model)

        self.setColumnWidth(0, 20)
        hheader.setSectionResizeMode(1, QHeaderView.ResizeToContents)

        self.log_view.instance.log.am_subscribe(self._on_new_logrecord)

    #
    # Public methods
    #

    def clear_log(self):
        self.log_view.instance.log.am_obj = [ ]
        self.log_view.instance.log.am_event()

    def copy_selected(self):
        content = [ ]
        selection = self.selectionModel().selectedRows()
        for row_index in selection:
            record = self.model.log[row_index.row()]
            content.append("%s | %s | %s | %s" % (
                QLogTableModel.level_to_text(record.level),
                str(record.timestamp),
                record.source,
                record.content
            ))
        self._copy_to_clipboard(os.linesep.join(content))

    def copy_selected_messages(self):
        content = [ ]
        selection = self.selectionModel().selectedRows()
        for row_index in selection:
            content.append(self.model.log[row_index.row()].content)
        self._copy_to_clipboard(os.linesep.join(content))

    def copy_all(self):
        content = [ ]
        for record in self.model.log:
            content.append("%s | %s | %s | %s" % (
                QLogTableModel.level_to_text(record.level),
                str(record.timestamp),
                record.source,
                record.content
            ))
        self._copy_to_clipboard(os.linesep.join(content))

    def copy_all_messages(self):
        content = [record.content for record in self.model.log]
        self._copy_to_clipboard(os.linesep.join(content))

    #
    # Events
    #

    def closeEvent(self, event):
        self.log_view.instance.log.am_unsubscribe(self._on_new_logrecord)
        super().closeEvent(event)

    def contextMenuEvent(self, arg__1: PySide6.QtGui.QContextMenuEvent):
        self._context_menu.popup(QCursor.pos())

    def _on_new_logrecord(self, log_record: LogRecord = None):
        gui_thread_schedule_async(self._on_new_logrecord_core, (log_record,))

    def _on_new_logrecord_core(self, log_record: LogRecord = None):
        self._before_row_insert()

        if log_record is None:
            # reload
            self.model.layoutAboutToBeChanged.emit()
            self.model._log = self.log_view.instance.log[::]
            self.model.layoutChanged.emit()
        else:
            log_records = len(self.model.log)
            self.model.rowsAboutToBeInserted.emit(self, log_records, log_records)
            self.model.log.append(log_record)
            self.model.rowsInserted.emit(self, log_records, log_records)

        self._after_row_insert()

    def _before_row_insert(self):
        scrollbar = self.verticalScrollBar()
        self._auto_scroll = scrollbar.value() == scrollbar.maximum()

    def _after_row_insert(self):
        if self._auto_scroll:
            self.scrollToBottom()

    def keyPressEvent(self, event:PySide6.QtGui.QKeyEvent):
        if event.matches(QKeySequence.Copy):
            self.copy_selected_messages()
        else:
            super().keyPressEvent(event)

    #
    # Private methods
    #

    @staticmethod
    def _copy_to_clipboard(content: str) -> None:
        clipboard = QGuiApplication.clipboard()
        clipboard.setText(content, QClipboard.Clipboard)
        if clipboard.supportsSelection():
            clipboard.setText(content, QClipboard.Selection)




from PySide6.QtGui import QKeySequence
from PySide6.QtCore import Qt

from .menu import Menu, MenuEntry, MenuSeparator


class AnalyzeMenu(Menu):
    def __init__(self, main_window):
        super().__init__("&Analyze", parent=main_window)
        self.entries.extend([
            MenuEntry('&Run Analysis...', main_window.run_analysis,
                      shortcut=QKeySequence(Qt.Key_F4)),
            MenuSeparator(),
            MenuEntry('&Decompile',
                main_window.decompile_current_function,
                shortcut=QKeySequence(Qt.Key_F5)),
            MenuEntry('View in Proximity &Browser',
                      main_window.view_proximity_for_current_function,
                      shortcut=QKeySequence(Qt.CTRL | Qt.Key_B)),
            MenuEntry('&Interact',
                main_window.interact,
                shortcut=QKeySequence(Qt.Key_F6)),
        ],)




def monkeypatch_rpyc():

    # The AsyncResult in rpyc has a bug that causes a race condition when clients are cascaded. this monkeypatch fixes
    # the bug.
    from rpyc.core.async_ import AsyncResult, AsyncResultTimeout

    def patched_wait(self):
        while not self._is_ready and not self._ttl.expired():
            self._conn.serve(0.01)
        if not self._is_ready:
            raise AsyncResultTimeout("result expired")

    AsyncResult.wait = patched_wait


monkeypatch_rpyc()


from .server import daemon_conn, daemon_exists, start_daemon, run_daemon_process
from .url_handler import handle_url



from typing import Mapping, Any, Sequence, Union

import angr


def extract_first_paragraph_from_docstring(desc: str) -> str:
    desc = desc.splitlines()
    last_line, first_line = -1, -1
    for idx, line in enumerate(desc):
        if first_line < 0:
            if len(line.strip()) > 0:
                first_line = idx
        else:
            if len(line.strip()) == 0:
                last_line = idx
                break

    if first_line >= 0:
        if last_line < 0:
            last_line = len(desc)
        desc = desc[first_line:last_line]
        num_whitespace_chars = len(desc[0]) - len(desc[0].lstrip())
        desc = ' '.join(l[num_whitespace_chars:] for l in desc)
    else:
        desc = ''

    return desc


class AnalysesConfiguration:
    """
    Configuration for a sequence of analyses.
    """

    def __init__(self, analyses: Sequence['AnalysisConfiguration'], instance):
        self.instance = instance
        self.analyses: Sequence['AnalysisConfiguration'] = analyses

    def __len__(self):
        return len(self.analyses)

    def __iter__(self):
        return iter(self.analyses)

    def __getitem__(self, key: Union[int, str]):
        if isinstance(key, int):
            return self.analyses[key]
        return self.by_name(key)

    def by_name(self, name: str) -> 'AnalysisConfiguration':
        for a in self.analyses:
            if a.name == name:
                return a
        raise KeyError(name)


class AnalysisConfiguration:
    """
    Configuration for an analysis.
    """

    def __init__(self, instance):
        self.instance = instance
        self.project: angr.Project = self.instance.project.am_obj
        self.enabled: bool = False
        self.name: str = ''
        self.display_name: str = ''
        self.description: str = 'Description not available'
        self.options: Mapping[str, AnalysisOption] = {}

    def __getitem__(self, key: str):
        return self.options[key]

    def to_dict(self):
        """
        Return dictionary with configuration for this option.
        """
        o = {}
        self.update_dict(o)
        return o

    def update_dict(self, out: Mapping[str, Any]):
        """
        Update dictionary `out` with configuration for this option.
        """
        for o in self.options.values():
            o.update_dict(out)


class AnalysisOption:
    """
    Configurable option for an analysis.
    """

    def __init__(self, name: str, display_name: str):
        self.name: str = name
        self.display_name: str = display_name

    def update_dict(self, out: Mapping[str, Any]):
        """
        Update dictionary `out` with configuration for this option.
        """


class PrimitiveAnalysisOption(AnalysisOption):
    """
    Configurable option for an analysis, with a fundamental type (e.g. bool)
    """

    def __init__(self, name: str, description: str, default: Any):
        super().__init__(name, description)
        self.default: Any = default
        self.value: Any = default

    def update_dict(self, out: Mapping[str, Any]):
        """
        Update `out` dictionary with configuration for this option.
        """
        out[self.name] = self.value


class BoolAnalysisOption(PrimitiveAnalysisOption):
    """
    Boolean option for an analysis.
    """

    def __init__(self, name: str, description: str, default: bool = False):
        super().__init__(name, description, default)


class CFGAnalysisConfiguration(AnalysisConfiguration):
    """
    Configuration for CFGFast analysis.
    """

    def __init__(self, instance):
        super().__init__(instance)
        self.name = 'cfg'
        self.display_name = 'Control-Flow Graph Recovery'
        self.description = extract_first_paragraph_from_docstring(self.project.analyses.CFGFast.__doc__)
        self.enabled = True
        self.options = {o.name: o for o in [
            BoolAnalysisOption('resolve_indirect_jumps', 'Resolve indirect jumps', True),
            BoolAnalysisOption('data_references', 'Collect cross-references and guess data types', True),
            BoolAnalysisOption('cross_references', 'Perform deep analysis on cross-references (slow)'),
            BoolAnalysisOption('skip_unmapped_addrs', 'Skip unmapped addresses', True),
            ]}


class FlirtAnalysisConfiguration(AnalysisConfiguration):
    """
    Configuration for Flirt analysis.
    """

    def __init__(self, instance):
        super().__init__(instance)
        self.name = 'flirt'
        self.display_name = 'Signature Matching'
        self.description = self.project.analyses.Flirt.__doc__.strip()
        self.enabled = True


class VariableRecoveryConfiguration(AnalysisConfiguration):
    """
    Configuration for VariableRecovery analysis.
    """

    def __init__(self, instance):
        super().__init__(instance)
        self.name = 'varec'
        self.display_name = 'Variable Recovery'
        self.description = extract_first_paragraph_from_docstring(self.project.analyses.VariableRecovery.__doc__)
        self.enabled = True
        self.options = {o.name: o for o in [
            BoolAnalysisOption('skip_signature_matched_functions', 'Skip variable recovery for signature-matched '
                                                                   'functions', True),
            ]}



import os.path

from PySide6.QtGui import QKeySequence
from PySide6.QtCore import Qt

from .menu import Menu, MenuEntry, MenuSeparator
from ...logic import GlobalInfo


class RecentMenuEntry(MenuEntry):
    """
    Represents an entry in the "Load recent" list. Holds a path and an indication of what's at that path.
    """
    def __init__(self, path):
        self.path = path
        super().__init__(os.path.basename(path), self.action_target)

    def action_target(self):
        GlobalInfo.main_window.load_file(self.path)

class FileMenu(Menu):
    """
    Lays out the entries under the 'File' menu
    """
    def __init__(self, main_window):
        super().__init__("&File", parent=main_window)

        self.recent_menu = Menu("Load recent")
        self.entries.extend([
            MenuEntry('L&oad a new binary...', main_window.open_file_button, shortcut=QKeySequence(Qt.CTRL | Qt.Key_O)),
            MenuEntry('Loa&d a new docker target...', main_window.open_docker_button,
                      shortcut=QKeySequence(Qt.SHIFT | Qt.CTRL | Qt.Key_O)),
            MenuEntry('Load a &trace file...', main_window.open_trace_file_button,
                      shortcut=QKeySequence(Qt.SHIFT | Qt.CTRL | Qt.Key_T)),
            self.recent_menu,
            MenuSeparator(),
            MenuEntry('&Load angr database...', main_window.load_database, shortcut=QKeySequence(Qt.CTRL | Qt.Key_L)),
            MenuEntry('&Save angr database...', main_window.save_database, shortcut=QKeySequence(Qt.CTRL | Qt.Key_S)),
            MenuEntry('S&ave angr database as...', main_window.save_database_as, shortcut=QKeySequence("Ctrl+Shift+S")),
            MenuSeparator(),
            MenuEntry('Load a new &trace...', main_window.load_trace),
            MenuSeparator(),
            MenuEntry('&Preferences...', main_window.preferences, shortcut=QKeySequence(Qt.CTRL | Qt.Key_P)),
            MenuSeparator(),
            MenuEntry('E&xit', main_window.quit),
        ])

    def add_recent(self, path: str):
        for entry in list(self.recent_menu.entries):
            if entry.path == path:
                self.recent_menu.remove(entry)
        self.recent_menu.add(RecentMenuEntry(path), 0)




import os
import atexit

from xdg import BaseDirectory

from .config_manager import ConfigurationManager

# Global configuration manager instance
fc = BaseDirectory.save_config_path('angr-management')
if fc is not None:
    config_path = os.path.join(fc, 'config')
    try:
        Conf = ConfigurationManager.parse_file(config_path)
    except FileNotFoundError:
        Conf = ConfigurationManager()
else:
    config_path = None
    print("Could not find configuration directory - settings will not be saved")
    Conf = ConfigurationManager()


def save_config():
    if fc is None:
        return
    Conf.save_file(config_path)


APP_LOCATION = str(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))
PLUGIN_PATH = str(os.path.join(APP_LOCATION, 'plugins'))
RES_LOCATION = str(os.path.join(APP_LOCATION, 'resources'))
IMG_LOCATION = str(os.path.join(RES_LOCATION, 'images'))
FONT_LOCATION = str(os.path.join(RES_LOCATION, 'fonts'))
THEME_LOCATION = str(os.path.join(RES_LOCATION, 'themes'))



"""
All functions in this module return a list of mixed BasePlugin class and exception objects to indicate
either the success or failure of a load.

This file intentionally does not reuse angr.misc.autoimport since it has additional design goals (managing exceptions,
loading out-of-tree files).
"""
import os
import logging
import importlib.util
from typing import List, Tuple

from .plugin_description import PluginDescription

l = logging.getLogger(__name__)


def load_plugin_descriptions_from_dir(path: str) -> List[Tuple[str,PluginDescription]]:
    try:
        dlist = os.listdir(path)
    except OSError:
        return [ ]

    plugins = [ ]
    for d in dlist:
        descs = load_plugin_description(os.path.join(path, d))
        for desc in descs:
            plugins.append((d, desc))
    return plugins


def load_plugin_description(path: str) -> List[PluginDescription]:
    try:
        flist = os.listdir(path)
    except OSError:
        return []

    if "plugin.toml" in flist:
        fpath = os.path.join(path, "plugin.toml")
        if os.path.isfile(fpath):
            return PluginDescription.from_toml(fpath)
    return []


def load_default_plugins():
    return load_plugins_from_dir(os.path.dirname(os.path.abspath(__file__)),
                                 exclude=('base_plugin.py', 'plugin_manager.py', 'plugin_description.py'))


def load_plugins_from_dir(path, exclude=()):
    out = []
    try:
        dlist = os.listdir(path)
    except OSError:
        return []

    for filename in dlist:
        if filename in exclude or filename in ('__init__.py', '__pycache__'):
            continue
        fullname = os.path.join(path, filename)
        if os.path.isfile(fullname) and fullname.endswith('.py'):
            out += load_plugins_from_file(fullname)
        elif os.path.isfile(os.path.join(fullname, '__init__.py')):
            out += load_plugins_from_package(fullname)

    return out


def load_plugins_from_package(path):
    # this logic is a little multiplexed in the next function but like... whatever
    return load_plugins_from_file(os.path.join(path, '__init__.py'))


def load_plugins_from_file(path):
    basename = os.path.basename(path)
    if basename == '__init__.py':
        modname = os.path.basename(os.path.dirname(path))
        if modname.count('.') != 0:
            l.error("file %s cannot be loaded - weird name", path)
            return []
    else:
        if os.path.isfile(path):
            modname = os.path.basename(os.path.dirname(path))
            modname += "." + basename.split('.')[0]
            if modname.count('.') != 1:
                l.error("package %s cannot be loaded - weird name", path)
                return []
        else:
            # directory
            modname = basename.split('.')[0]
            if basename.count('.') != 1:
                l.error("package %s cannot be loaded - weird name", path)
                return []

    # https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
    spec = importlib.util.spec_from_file_location("angrmanagement.plugins.%s" % modname, path)
    if spec is None:
        l.error("Not a python module: %s", path)
        return []
    try:
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
    except Exception as e:
        return [e]

    return load_plugins_from_module(mod)


def load_plugins_from_module(module):
    return load_plugins_from_vars(vars(module))


def load_plugins_from_vars(variables):
    out = []
    for _, cls in variables.items():
        if type(cls) is type and issubclass(cls, BasePlugin) and not hasattr(cls, '_%s__i_hold_this_abstraction_token' % cls.__name__):
            out.append(cls)
    return out


from .plugin_manager import PluginManager
from .base_plugin import BasePlugin



from typing import Type

from PySide6.QtGui import QKeySequence
from angrmanagement.ui.toolbars.toolbar import Toolbar

from .menu import Menu, MenuEntry, MenuSeparator


class NewViewMenu(Menu):
    """
    Sub-menu to construct new Views
    """

    def __init__(self, main_window):
        super().__init__("&New", parent=main_window)

        self.entries.extend([
            MenuEntry('&Linear Disassembly', main_window.workspace.create_and_show_linear_disassembly_view),
            MenuEntry('&Graph Disassembly', main_window.workspace.create_and_show_graph_disassembly_view,
                      shortcut=QKeySequence("Ctrl+N")),
            MenuSeparator(),
            MenuEntry('&Hex', main_window.workspace.create_and_show_hex_view),
        ])


class ToolbarMenuEntry(MenuEntry):
    """
    Menu item to control toolbar visibility.
    """

    def __init__(self, toolbar_cls: Type[Toolbar], main_window: 'MainWindow'):
        tm = main_window.toolbar_manager
        super().__init__(tm.get_name_for_toolbar_class(toolbar_cls), self.on_toggle, checkable=True)
        self.main_window = main_window
        self.toolbar_cls = toolbar_cls

    @property
    def is_visibile(self) -> bool:
        return self.toolbar_cls in self.main_window.toolbar_manager.active

    def on_toggle(self):
        self.main_window.toolbar_manager.set_toolbar_visible_by_class(self.toolbar_cls, not self.is_visibile)

    def update_checked(self):
        self.qaction.setChecked(self.is_visibile)


class ToolbarMenu(Menu):
    """
    Sub-menu to control toolbar visibility.
    """

    def __init__(self, main_window):
        super().__init__("&Toolbars", parent=main_window)
        tm = main_window.toolbar_manager
        self.entries.extend([MenuEntry('Show all', tm.show_all), MenuEntry('Hide all', tm.hide_all), MenuSeparator()])
        for tb_cls in tm.all_toolbars:
            self.entries.append(ToolbarMenuEntry(tb_cls, main_window))
        self.qmenu().aboutToShow.connect(self.update_checked_entries)

    def update_checked_entries(self):
        for e in self.entries:
            if isinstance(e, ToolbarMenuEntry):
                e.update_checked()


class ViewMenu(Menu):
    """
    Main View menu
    """

    def __init__(self, main_window):
        super().__init__("&View", parent=main_window)

        self.entries.extend([
            ToolbarMenu(main_window),
            MenuSeparator(),
            MenuEntry('Next Tab', main_window.workspace.view_manager.next_tab, shortcut=QKeySequence("Ctrl+Tab")),
            MenuEntry('Previous Tab', main_window.workspace.view_manager.previous_tab,
                      shortcut=QKeySequence("Ctrl+Shift+Tab")),
            MenuSeparator(),
            NewViewMenu(main_window),
            MenuSeparator(),
            MenuEntry('&Linear Disassembly', main_window.workspace.show_linear_disassembly_view),
            MenuEntry('&Graph Disassembly', main_window.workspace.show_graph_disassembly_view),
            MenuSeparator(),
            MenuEntry('&Hex', main_window.workspace.show_hex_view),
            MenuEntry('Pro&ximity', main_window.view_proximity_for_current_function),
            MenuEntry('Pseudo&code', main_window.workspace.show_pseudocode_view),
            MenuEntry('&Strings', main_window.workspace.show_strings_view),
            MenuEntry('&Patches', main_window.workspace.show_patches_view),
            MenuEntry('&Types', main_window.workspace.show_types_view),
            MenuEntry('&Functions', main_window.workspace.show_functions_view),
            MenuEntry('&Traces', main_window.workspace.show_traces_view),
            MenuEntry('&Trace Map', main_window.workspace.show_trace_map_view),
            MenuSeparator(),
            MenuEntry('Symbolic &Execution', main_window.workspace.show_symexec_view),
            MenuEntry('S&ymbolic States', main_window.workspace.show_states_view),
            MenuEntry('&Interaction', main_window.workspace.show_interaction_view),
            MenuEntry('&Registers', main_window.workspace.show_registers_view),
            MenuEntry('&Stack', main_window.workspace.show_stack_view),
            MenuEntry('&Breakpoints', main_window.workspace.show_breakpoints_view),
            MenuEntry('&Call Explorer', main_window.workspace.show_call_explorer_view),
            MenuSeparator(),
            MenuEntry('&Console', main_window.workspace.show_console_view),
            MenuEntry('&Log', main_window.workspace.show_log_view),
        ])



from .cfg_generation import CFGGenerationJob
from .code_tagging import CodeTaggingJob
from .ddg_generation import DDGGenerationJob
from .simgr_explore import SimgrExploreJob
from .simgr_step import SimgrStepJob
from .vfg_generation import VFGGenerationJob
from .prototype_finding import PrototypeFindingJob
from .variable_recovery import VariableRecoveryJob
from .dependency_analysis import DependencyAnalysisJob
from .decompile_function import DecompileFunctionJob
from .flirt_signature_recognition import FlirtSignatureRecognitionJob



import logging
import os
from typing import Optional

import angr
import binsync
from angr.analyses.decompiler.structured_codegen import DummyStructuredCodeGenerator
from angrmanagement.ui.views import CodeView
from binsync.common.controller import (
    BinSyncController,
    init_checker,
    make_ro_state,
    make_state_with_func
)
from binsync.data import FunctionHeader, StackOffsetType

l = logging.getLogger(__name__)


class AngrBinSyncController(BinSyncController):
    """
    The class used for all pushing/pulling and merging based actions with BinSync data.
    This class is resposible for handling callbacks that are done by changes from the local user
    and responsible for running a thread to get new changes from other users.
    """

    def __init__(self, workspace):
        super().__init__()
        self._workspace = workspace
        self._instance = workspace.main_instance

    def binary_hash(self) -> str:
        return self._instance.project.loader.main_object.md5.hex()

    def active_context(self):
        curr_view = self._workspace.view_manager.current_tab
        if not curr_view:
            return None

        try:
            func = curr_view.function
        except NotImplementedError:
            return None

        if func is None or func.am_obj is None:
            return None

        func_addr = self.rebase_addr(func.addr)

        return binsync.data.Function(
            func_addr, 0, header=FunctionHeader(func.name, func_addr)
        )

    def binary_path(self) -> Optional[str]:
        try:
            return self._instance.project.loader.main_object.binary
        # pylint: disable=broad-except
        except Exception:
            return None

    def get_func_size(self, func_addr) -> int:
        try:
            func = self._instance.kb.functions[func_addr]
            return func.size
        except KeyError:
            return 0

    def rebase_addr(self, addr, up=False):
        base_addr = self._instance.project.loader.main_object.mapped_base
        is_pie = self._instance.project.loader.main_object.pic

        if is_pie:
            if up:
                return addr + base_addr
            elif addr > base_addr:
                return addr - base_addr

        return addr

    #
    # Display Fillers
    #

    def fill_struct(self, struct_name, user=None, state=None):
        pass

    @init_checker
    @make_ro_state
    def fill_function(self, func_addr, user=None, state=None):
        func = self._instance.kb.functions[self.rebase_addr(func_addr, up=True)]

        # re-decompile a function if needed
        decompilation = self.decompile_function(func)

        sync_func: binsync.data.Function = self.pull_function(self.rebase_addr(func.addr), user=user)
        if sync_func is None:
            # the function does not exist for that user's state
            return False

        sync_func = self.generate_func_for_sync_level(sync_func)

        #
        # Function Header
        #

        if sync_func.header:
            if sync_func.name and sync_func.name != func.name:
                func.name = sync_func.name
                decompilation.cfunc.name = sync_func.name
                decompilation.cfunc.demangled_name = sync_func.name

            if sync_func.header.args:
                for i, arg in sync_func.header.args.items():
                    if i >= len(decompilation.cfunc.arg_list):
                        break

                    decompilation.cfunc.arg_list[i].variable.name = arg.name

        #
        # Comments
        #

        for addr, cmt in self.pull_func_comments(self.rebase_addr(func_addr)).items():
            if not cmt or not cmt.comment:
                continue

            if cmt.decompiled:
                try:
                    pos = decompilation.map_addr_to_pos.get_nearest_pos(addr)
                    corrected_addr = decompilation.map_pos_to_addr.get_node(pos).tags['ins_addr']
                # pylint: disable=broad-except
                except Exception:
                    break

                decompilation.stmt_comments[corrected_addr] = cmt.comment
            else:
                self._instance.kb.comments[cmt.addr] = cmt.comment

        # ==== Stack Vars ==== #
        sync_vars = self.pull_stack_variables(self.rebase_addr(func.addr), user=user)
        for offset, sync_var in sync_vars.items():
            code_var = AngrBinSyncController.find_stack_var_in_codegen(decompilation, offset)
            if code_var:
                code_var.name = sync_var.name
                code_var.renamed = True

        decompilation.regenerate_text()
        self.decompile_function(func, refresh_gui=True)
        return True

    #
    #   Pushers
    #

    @init_checker
    @make_state_with_func
    # pylint: disable=arguments-differ
    def push_function_header(self, addr, new_name, ret_type=None, args=None, user=None, state=None):
        func_header = FunctionHeader(new_name, addr, ret_type=ret_type, args=args)
        return state.set_function_header(func_header)

    @init_checker
    @make_state_with_func
    # pylint: disable=arguments-differ
    def push_stack_variable(self, func_addr, offset, name, type_, size_, user=None, state=None):
        sync_var = binsync.data.StackVariable(offset, StackOffsetType.ANGR, name, type_, size_, func_addr)
        return state.set_stack_variable(sync_var, offset, func_addr)

    @init_checker
    @make_state_with_func
    # pylint: disable=unused-argument,arguments-differ
    def push_comment(self, addr, cmt, decompiled, func_addr=None, user=None, state=None):
        sync_cmt = binsync.data.Comment(addr, cmt, decompiled=decompiled)
        return state.set_comment(sync_cmt)

    #
    #   Utils
    #

    def decompile_function(self, func, refresh_gui=False):
        # check for known decompilation
        available = self._instance.kb.structured_code.available_flavors(func.addr)
        should_decompile = False
        if 'pseudocode' not in available:
            should_decompile = True
        else:
            cached = self._instance.kb.structured_code[(func.addr, 'pseudocode')]
            if isinstance(cached, DummyStructuredCodeGenerator):
                should_decompile = True

        if should_decompile:
            # recover direct pseudocode
            self._instance.project.analyses.Decompiler(func, flavor='pseudocode')

            # attempt to get source code if its available
            source_root = None
            if self._instance.original_binary_path:
                source_root = os.path.dirname(self._instance.original_binary_path)
            self._instance.project.analyses.ImportSourceCode(func, flavor='source', source_root=source_root)

        # grab newly cached pseudocode
        decomp = self._instance.kb.structured_code[(func.addr, 'pseudocode')].codegen
        if refresh_gui:
            # refresh all views
            self._workspace.reload()

            # re-decompile current view to cause a refresh
            current_tab = self._workspace.view_manager.current_tab
            if isinstance(current_tab, CodeView) and current_tab.function == func:
                self._workspace.decompile_current_function()

        return decomp

    @staticmethod
    def find_stack_var_in_codegen(decompilation, stack_offset: int) -> Optional[angr.sim_variable.SimStackVariable]:
        for var in decompilation.cfunc.variable_manager._unified_variables:
            if hasattr(var, "offset") and var.offset == stack_offset:
                return var

        return None

    @staticmethod
    def stack_var_type_str(decompilation, stack_var: angr.sim_variable.SimStackVariable):
        try:
            var_type = decompilation.cfunc.variable_manager.get_variable_type(stack_var)
        # pylint: disable=broad-except
        except Exception:
            return None

        return var_type.c_repr()

    @staticmethod
    def get_func_args(decompilation):
        arg_info = {
            i: (arg.variable, decompilation.cfunc.functy.args[i].c_repr())
            for i, arg in enumerate(decompilation.cfunc.arg_list)
        }
        return arg_info

    @staticmethod
    def func_insn_addrs(func: angr.knowledge_plugins.Function):
        insn_addrs = set()
        for block in func.blocks:
            insn_addrs.update(block.instruction_addrs)

        return insn_addrs

    def get_func_addr_from_addr(self, addr):
        try:
            func_addr = self._workspace.main_instance.kb.cfgs.get_most_accurate()\
                .get_any_node(addr, anyaddr=True)\
                .function_address
        except AttributeError:
            func_addr = None

        return func_addr
    
    def goto_address(self, func_addr):
        self._workspace.jump_to(self.rebase_addr(func_addr, up=True))



from typing import List, Optional


class ComponentFunction:

    __slots__ = ('mapped_base', 'virtual_addr', 'symbol_name', )

    def __init__(self, mapped_base: int, virtual_addr: int, symbol_name: Optional[str]=None):
        self.mapped_base = mapped_base
        self.virtual_addr = virtual_addr
        self.symbol_name = symbol_name

    def __eq__(self, other):
        return isinstance(other, ComponentFunction) and \
                self.mapped_base == other.mapped_base and \
                self.virtual_addr == other.virtual_addr

    def __hash__(self):
        return hash((ComponentFunction, self.mapped_base, self.virtual_addr))


class ComponentTreeNode:
    def __init__(self, name=None):
        self.name = name
        self.components: List['ComponentTreeNode'] = [ ]
        self.functions: List[ComponentFunction] = [ ]

    def __eq__(self, other):
        return isinstance(other, ComponentTreeNode) \
                and self.components == other.components \
                and set(self.functions) == set(other.functions)

    def __hash__(self):
        return hash((ComponentTreeNode,
                     hash(tuple(self.components)),
                     hash(tuple(sorted((f.mapped_base + f.virtual_addr) for f in self.functions))),
                     )
                    )


class ComponentTree:
    def __init__(self, root: Optional[ComponentTreeNode]=None):
        self.root = root



from PySide6.QtWidgets import QTableWidget, QTableWidgetItem, QAbstractItemView, QMenu, QFileDialog, QHeaderView


class QFileSystemTable(QTableWidget):
    def __init__(self, items, parent):
        super().__init__(parent)

        header_labels = ['Mount Point', 'Host Path']

        self.setColumnCount(len(header_labels))
        self.setHorizontalHeaderLabels(header_labels)
        self.setSelectionBehavior(QAbstractItemView.SelectItems)
        header = self.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.Stretch)
        header.setSectionResizeMode(1, QHeaderView.Stretch)


        self.setRowCount(len(items))
        for idx, item in enumerate(items):
            for i, it in enumerate(item):
                self.setItem(idx, i, QTableWidgetItem(it))

    def contextMenuEvent(self, event):
        sr = self.currentRow()

        menu = QMenu("", self)

        menu.addAction('Add a Row', self._action_new_row)
        menu.addSeparator()

        a = menu.addAction('Delete this Row', self._action_delete)
        if sr is None:
            a.setDisabled(True)
        b = menu.addAction('Select a File', self._action_select_file)
        if sr is None:
            b.setDisabled(True)
        c = menu.addAction('Select a directory', self._action_select_dir)
        if sr is None:
            c.setDisabled(True)


        menu.exec_(event.globalPos())

    def _action_new_row(self):
        row = self.rowCount()
        self.insertRow(row)
        self.setItem(row, 0, QTableWidgetItem("Edit Me"))
        self.setItem(row, 1, QTableWidgetItem(""))

    def _action_select_file(self):
        file_path, succ = QFileDialog.getOpenFileName(self, "Open a real file", "","All executables (*)",)
        if succ:
            self.setItem(self.currentRow(),1,QTableWidgetItem(file_path))

    def _action_select_dir(self):
        dir_path = QFileDialog.getExistingDirectory(self, "Select a directory")
        if dir_path:
            self.setItem(self.currentRow(),1,QTableWidgetItem(dir_path))

    def _action_delete(self):
        self.removeRow(self.currentRow())

    def get_result(self):
        ret = []
        for i in range(self.rowCount()):
            ret.append([self.item(i,0).text(), self.item(i,1).text()])
        return ret




import re

from pyqodeng.core.api import SyntaxHighlighter
from PySide6.QtGui import QTextCharFormat, QFont, QBrush

from ..documents import QCodeDocument
from ...config import Conf

FORMATS = {}

def reset_formats():
    f = QTextCharFormat()
    f.setForeground(QBrush(Conf.pseudocode_keyword_color))
    f.setFontWeight(QFont.Bold)
    FORMATS['keyword'] = f

    f = QTextCharFormat()
    f.setForeground(QBrush(Conf.pseudocode_quotation_color))
    FORMATS['quotation'] = f

    f = QTextCharFormat()
    f.setForeground(QBrush(Conf.pseudocode_function_color))
    f.setFontWeight(QFont.Bold)
    FORMATS['function'] = f

    f = QTextCharFormat()
    f.setForeground(QBrush(Conf.pseudocode_comment_color))
    f.setFontWeight(QFont.Bold)
    FORMATS['comment'] = f

reset_formats()

class QCCodeHighlighter(SyntaxHighlighter):
    """
    A syntax highlighter for QCCodeEdit. Uses a custom lexing scheme to detect C constructs (functions, keywords,
    comments, and strings) and adds styling to them based on the current color scheme.
    """

    HIGHLIGHTING_RULES = [
        # quotation
        #(r"\"([^\\\"]|(\\.))*\"", 'quotation'),
        # comment
        #(r"//[^\n]*", 'comment'),
        #(r"/\*[^\n]*\*/", 'comment'),
        # function
        (r"\b[A-Za-z0-9_:]+\s*(?=\()", 'function'),
        (r"\bNULL\b", 'function'),
        # keywords
        (r"\bbool\b", 'keyword'),
        (r"\bbreak\b", 'keyword'),
        (r"\bcase\b", 'keyword'),
        (r"\bcatch\b", 'keyword'),
        (r"\bchar\b", 'keyword'),
        (r"\bclass\b", 'keyword'),
        (r"\bconst\b", 'keyword'),
        (r"\bcontinue\b", 'keyword'),
        (r"\bdefault\b", 'keyword'),
        (r"\bdelete\b", 'keyword'),
        (r"\bdo\b", 'keyword'),
        (r"\bdouble\b", 'keyword'),
        (r"\belse\b", 'keyword'),
        (r"\benum\b", 'keyword'),
        (r"\bexplicit\b", 'keyword'),
        (r"\bfloat\b", 'keyword'),
        (r"\bfor\b", 'keyword'),
        (r"\bfriend\b", 'keyword'),
        (r"\bgoto\b", 'keyword'),
        (r"\bif\b", 'keyword'),
        (r"\binline\b", 'keyword'),
        (r"\bint\b", 'keyword'),
        (r"\blong\b", 'keyword'),
        (r"\bnamespace\b", 'keyword'),
        (r"\bnew\b", 'keyword'),
        (r"\boperator\b", 'keyword'),
        (r"\bprivate\b", 'keyword'),
        (r"\bprotected\b", 'keyword'),
        (r"\bpublic\b", 'keyword'),
        (r"\bshort\b", 'keyword'),
        (r"\bsigned\b", 'keyword'),
        (r"\bsizeof\b", 'keyword'),
        (r"\bstatic\b", 'keyword'),
        (r"\bstruct\b", 'keyword'),
        (r"\btemplate\b", 'keyword'),
        (r"\bthis\b", 'keyword'),
        (r"\btrue\b", 'keyword'),
        (r"\btypedef\b", 'keyword'),
        (r"\btypename\b", 'keyword'),
        (r"\bunion\b", 'keyword'),
        (r"\bunsigned\b", 'keyword'),
        (r"\bvirtual\b", 'keyword'),
        (r"\bvoid\b", 'keyword'),
        (r"\bvolatile\b", 'keyword'),
        (r"\bwhile\b", 'keyword'),
        (r"\bswitch\b", 'keyword'),
        (r"\breturn\b", 'keyword'),
    ]

    def __init__(self, parent, color_scheme=None):
        super().__init__(parent, color_scheme=color_scheme)

        self.doc = parent  # type: QCodeDocument
        self.comment_status = False

    def highlight_block(self, text, block):
        # this code makes the assumption that this function is only ever called on lines in sequence in order
        # it might also fuck up if it ever calls it starting in the middle...
        if block.previous() is None:
            self.comment_status = False

        mark_in = 0

        quote_status = False
        quote_mark = None
        escape_counter = 0
        for col, _ in enumerate(text):
            if quote_status:
                assert not self.comment_status
                if escape_counter:
                    escape_counter -= 1
                elif text[col] == '\\':
                    escape_counter = 1
                elif text[col] == quote_mark:
                    quote_status = False
                    mark_out = col + 1
                    self.setFormat(mark_in, mark_out - mark_in, FORMATS['quotation'])
                    text = text[:mark_in] + " " * (mark_out - mark_in) + text[mark_out:]
            else:
                if self.comment_status:
                    if text[col:col+2] == "*/":
                        mark_out = col + 2
                        self.comment_status = False
                        self.setFormat(mark_in, mark_out - mark_in, FORMATS['comment'])
                        text = text[:mark_in] + " " * (mark_out - mark_in) + text[mark_out:]
                else:
                    # regular mode is here
                    if text[col] in ('"', "'"):
                        mark_in = col
                        quote_status = True
                        quote_mark = text[col]
                    elif text[col:col+2] == "/*":
                        mark_in = col
                        self.comment_status = True
                    elif text[col:col+2] == '//':
                        # do not set comment_status. just format the line and break.
                        mark_in = col
                        mark_out = len(text)
                        self.setFormat(mark_in, mark_out - mark_in, FORMATS['comment'])
                        text = text[:mark_in] + " " * (mark_out - mark_in) + text[mark_out:]
                        break
        if self.comment_status:
            # do not unset. this is a multiline comment
            mark_out = len(text)
            self.setFormat(mark_in, mark_out - mark_in, FORMATS['comment'])
            text = text[:mark_in] + " " * (mark_out - mark_in) + text[mark_out:]

        for pattern, format_id in self.HIGHLIGHTING_RULES:
            for mo in list(re.finditer(pattern, text)):
                start = mo.start()
                end = mo.end()
                self.setFormat(start, end - start, FORMATS[format_id])
                #if format_id in {'quotation', 'comment'}:
                #    # remove the formatted parts so that we do not end up highlighting these parts again
                #    text = text[:start] + " " * (end-start) + text[end:]



from typing import Type, Mapping

from PySide6.QtCore import Qt
from angrmanagement.ui.toolbars import FileToolbar, DebugToolbar
from angrmanagement.ui.toolbars.toolbar import Toolbar


class ToolbarManager:
    """
    Manages toolbars shown on the main window.
    """

    def __init__(self, main_window):
        self._main_window: 'MainWindow' = main_window
        self.active: Mapping[Type[Toolbar], Toolbar] = {}
        self.all_toolbars = [FileToolbar, DebugToolbar]

    @staticmethod
    def get_name_for_toolbar_class(toolbar_cls: Type[Toolbar]) -> str:
        return {
            FileToolbar: 'File',
            DebugToolbar: 'Debug'
        }[toolbar_cls]

    def show_toolbar_by_class(self, cls: Type[Toolbar]):
        if cls not in self.active:
            tb = cls(self._main_window)
            self.active[cls] = tb
            self._main_window.addToolBar(Qt.TopToolBarArea, tb.qtoolbar())

    def hide_toolbar_by_class(self, cls: Type[Toolbar]):
        if cls in self.active:
            tb = self.active.pop(cls)
            self._main_window.removeToolBar(tb.qtoolbar())

    def set_toolbar_visible_by_class(self, cls: Type[Toolbar], visible: bool):
        if visible:
            self.show_toolbar_by_class(cls)
        else:
            self.hide_toolbar_by_class(cls)

    def show_all(self):
        for cls in self.all_toolbars:
            self.show_toolbar_by_class(cls)

    def hide_all(self):
        for cls in self.all_toolbars:
            self.hide_toolbar_by_class(cls)




from PySide6.QtGui import QKeySequence
from PySide6.QtCore import Qt

from .menu import Menu, MenuEntry, MenuSeparator


class HelpMenu(Menu):
    def __init__(self, main_window):
        super().__init__("&Help", parent=main_window)

        self.entries.extend([
            MenuEntry('&Documentation', main_window.open_doc_link,  shortcut=QKeySequence(Qt.ALT | Qt.Key_H)),
            MenuSeparator(),
            MenuEntry('About angr...', main_window.open_about_dialog)
        ])



import time

from PySide6.QtCore import Qt, QAbstractTableModel, QModelIndex, QEvent, Signal, QObject
from PySide6.QtGui import QCursor
from PySide6.QtWidgets import (
    QVBoxLayout,
    QMainWindow,
    QTableView,
    QAbstractItemView,
    QHeaderView,
    QWidget,
    QHBoxLayout,
    QComboBox,
    QLabel,
    QPushButton, QLineEdit, QCheckBox, QMenu, QFileDialog,
)

from angrmanagement.plugins import BasePlugin
from angrmanagement.ui.views import BaseView
from angrmanagement.ui.workspace import Workspace
import codecs

from .seed_table import SeedTable

class querySignaler(QObject):
    querySignal = Signal(bool)

class SeedTableModel(QAbstractTableModel):

    def __init__(self, workspace, table, dropdown, countlabel):
        super(SeedTableModel, self).__init__()
        self.query_signal = querySignaler()
        self.query_signal.querySignal.connect(self.querySignalHandle)
        self.seed_db = SeedTable(workspace, self.query_signal, seed_callback=self.add_seed)
        self.countlabel = countlabel
        self.table = table
        self.workspace = workspace
        self.page_dropdown = dropdown
        self.headers = ["ID", "Input", "NC", "C", "NT", "L", "E"]
        self.seeds = []
        self.displayed_seeds = []

        # pagination support
        self.current_page = 1
        self.max_pages = 1
        self.entries_per_page = 50

        self.set_page(1)

    def rowCount(self, index=QModelIndex()):
        if not self.displayed_seeds:
            return 0
        return len(self.displayed_seeds)

    def columnCount(self, index=QModelIndex()):
        return len(self.headers)

    def querySignalHandle(self, status):
        if status:
            self.countlabel.setText("<font color=#ff0000>Querying..</font>")
            self.countlabel.repaint()
    # probably not useful anymore. kept for not wanting to do it again.
    # def canFetchMore(self, index=QModelIndex()):
    #     return len(self.seeds) > self.num_loaded
    #
    # def fetchMore(self, index=QModelIndex()):
    #     num_to_fetch = min(len(self.seeds) - self.num_loaded, 50)
    #     self.beginInsertRows(QModelIndex(), self.num_loaded, self.num_loaded+num_to_fetch-1)
    #     self.num_loaded += num_to_fetch
    #     self.endInsertRows()
    #     self.table.resizeEvent(QResizeEvent(self.table.size(), QSize()))

    def set_page(self, pagenum):
        self.beginResetModel()
        if self.max_pages >= pagenum > 0:
            self.current_page = pagenum
        else:
            return False
        # load seeds for page
        min_index = (pagenum - 1) * self.entries_per_page
        max_index = min((pagenum * self.entries_per_page) - 1, len(self.seeds))
        # check to ensure we arent out of bounds
        if min_index > len(self.seeds):  # this should REALLY never happen.
            print("ERROR: Invalid page selected.")
            return False
        self.displayed_seeds = self.seeds[min_index:max_index]
        self.endResetModel()
        return True

    def add_seed(self, seed):
        self.beginResetModel()
        # more complex logic here.. probably
        if isinstance(seed, list):
            for s in seed:
                self.seeds.append(s)
        else:
            self.seeds.append(seed)
        # update our page
        self.max_pages = max(len(self.seeds) // self.entries_per_page, 1)
        self.set_page(self.current_page)
        self.page_dropdown.clear()
        self.page_dropdown.addItems(list(map(str, range(1, self.max_pages+1))))
        self.countlabel.setText("Count: " + str(len(self.seeds)))
        self.endResetModel()

    def clear_seeds(self):
        self.beginResetModel()
        self.seeds = []
        self.displayed_seeds = []
        self.max_pages = max(len(self.seeds) // self.entries_per_page, 1)
        self.page_dropdown.clear()
        self.page_dropdown.addItems(list(map(str, range(1, self.max_pages+1))))
        self.countlabel.setText("Count: " + str(len(self.seeds)))
        self.endResetModel()
        self.set_page(1)

    def data(self, index, role=Qt.DisplayRole):
        col = index.column()
        seed = self.displayed_seeds[index.row()]
        if role == Qt.DisplayRole:
            if col == 0:
                return seed.id
            elif col == 1:
                return repr(seed.value) if len(seed.value) < 80 else repr(seed.value[:80] + b"...")
            elif col == 2 and "non-crashing" in seed.tags:
                return "x"
            elif col == 3 and "crashing" in seed.tags:
                return "x"
            elif col == 4 and "non-terminating" in seed.tags:
                return "x"
            elif col == 5 and "leaking" in seed.tags:
                return "x"
            elif col == 6 and "exploit" in seed.tags:
                return "x"
            return None
        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role != Qt.DisplayRole:
            return None
        if orientation == Qt.Horizontal and section < len(self.headers):
            return self.headers[section]
        else:
            return None

    def go_next_page(self):
        if self.set_page(self.current_page + 1):
            self.page_dropdown.clear()
            self.page_dropdown.addItems(list(map(str, range(1, self.max_pages+1))))
            self.page_dropdown.setCurrentIndex(self.current_page - 1)

    def go_prev_page(self):
        if self.set_page(self.current_page - 1):
            self.page_dropdown.clear()
            self.page_dropdown.addItems(list(map(str, range(1, self.max_pages+1))))
            self.page_dropdown.setCurrentIndex(self.current_page - 1)


class SeedTableWidget(QTableView):
    def __init__(self, parent, workspace):
        super().__init__(parent)
        self.workspace = workspace

    def refresh(self):
        self.viewport().update()

    def init_parameters(self):
        self.horizontalHeader().setVisible(True)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setDefaultSectionSize(18)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)

    def contextMenuEvent(self, event:'PySide6.QtGui.QContextMenuEvent') -> None:
        rows = self.selectionModel().selectedIndexes()
        contextMenu = QMenu(self)
        saveSeed = contextMenu.addAction("&Save Seed")
        action = contextMenu.exec_(QCursor.pos())
        if action == saveSeed:
            self.saveSeed(rows)

    def saveSeed(self, rows):
        data = self.model().displayed_seeds[rows[0].row()].value
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        filename = QFileDialog.getSaveFileName(self, "Save Seed", "", "All Files(*)", options=options)[0]
        try:
            with open(filename, "wb") as outfile:
                outfile.write(data)
        except:
            self.workspace.log("Error saving seed.")



class SeedTableView(BaseView):
    def __init__(self, workspace: Workspace, *args, **kwargs):
        super().__init__("SeedTableView", workspace, *args, **kwargs)
        self.base_caption = "Seed Table"
        self.workspace = workspace
        self.instance = workspace.main_instance
        workspace.main_instance.project.am_subscribe(self.on_project_load)
        self._init_widgets()

    def page_changed(self, i):
        self.table_data.set_page(self.page_dropdown.currentIndex() + 1)

    def _init_widgets(self):
        self.main = QMainWindow()
        self.main.setWindowFlags(Qt.Widget)

        self.container = QWidget()  # create containing widget to keep things nice
        self.container.setLayout(QVBoxLayout())

        # count label
        self.seed_count_label = QLabel("Count:")

        # create table
        self.page_dropdown = QComboBox()
        self.table = SeedTableWidget(self, self.workspace)
        self.table_data = SeedTableModel(self.workspace, self.table, self.page_dropdown, self.seed_count_label)
        self.table.setModel(self.table_data)
        self.table.init_parameters()  # need to set table model before messing with column resizing
        self.container.layout().addWidget(self.table)

        # create bottom section
        self.bottom_widget = QWidget()
        self.bottom_widget.setLayout(QHBoxLayout())
        # page buttons
        self.next_page_btn = QPushButton(">")
        self.next_page_btn.setMaximumWidth(40)
        self.next_page_btn.clicked.connect(self.table_data.go_next_page)
        self.prev_page_btn = QPushButton("<")
        self.prev_page_btn.setMaximumWidth(40)
        self.prev_page_btn.clicked.connect(self.table_data.go_prev_page)
        # page label
        self.page_label = QLabel("Page:")
        # page dropdown
        self.page_dropdown.addItems(list(map(str, range(1, 1))))  # test
        self.page_dropdown.setCurrentIndex(0)
        self.page_dropdown.activated.connect(self.page_changed)
        # filter box
        self.filter_box = SeedTableFilterBox(self)
        self.filter_box.returnPressed.connect(self._on_filter_change)
        # filter checkboxes
        # "NC", "C", "NT", "L", "E"
        self.nc_checkbox = QCheckBox("NC")
        self.nc_checkbox.stateChanged.connect(self._on_filter_change)
        self.c_checkbox = QCheckBox("C")
        self.c_checkbox.stateChanged.connect(self._on_filter_change)
        self.nt_checkbox = QCheckBox("NT")
        self.nt_checkbox.stateChanged.connect(self._on_filter_change)
        self.l_checkbox = QCheckBox("L")
        self.l_checkbox.stateChanged.connect(self._on_filter_change)
        self.e_checkbox = QCheckBox("E")
        self.e_checkbox.stateChanged.connect(self._on_filter_change)


        self.bottom_widget.layout().addWidget(self.seed_count_label)
        self.bottom_widget.layout().addWidget(self.filter_box)
        self.bottom_widget.layout().addWidget(self.nc_checkbox)
        self.bottom_widget.layout().addWidget(self.c_checkbox)
        self.bottom_widget.layout().addWidget(self.nt_checkbox)
        self.bottom_widget.layout().addWidget(self.l_checkbox)
        self.bottom_widget.layout().addWidget(self.e_checkbox)
        # self.bottom_widget.layout().addStretch()
        self.bottom_widget.layout().addWidget(self.prev_page_btn)
        self.bottom_widget.layout().addWidget(self.page_label)
        self.bottom_widget.layout().addWidget(self.page_dropdown)
        self.bottom_widget.layout().addWidget(self.next_page_btn)

        self.container.layout().addWidget(self.bottom_widget)

        self.main.setCentralWidget(self.container)
        main_layout = QVBoxLayout()
        main_layout.addWidget(self.main)
        self.setLayout(main_layout)

    def on_project_load(self, **kwargs):
        if self.instance.project.am_none:
            return
        pass

    def _on_filter_change(self):
        raw_filter = self.filter_box.text()
        inp = None
        if len(raw_filter) > 0:
            inp, _ = codecs.escape_decode(raw_filter, 'hex')
        flags = []
        if self.nc_checkbox.isChecked():
            flags.append("non-crashing")
        if self.c_checkbox.isChecked():
            flags.append("crashing")
        if self.l_checkbox.isChecked():
            flags.append("leaking")
        if self.nt_checkbox.isChecked():
            flags.append("non-terminating")
        if self.e_checkbox.isChecked():
            flags.append("exploit")

        self.table_data.clear_seeds()
        if len(flags) == 0 and inp is None:
            data = self.table_data.seed_db.get_all_seeds()
        else:
            if inp:
                data = self.table_data.seed_db.filter_seeds_by_value(inp)
                data = list(filter(lambda s: all([x in s.tags for x in flags]), data))
            else:
                data = self.table_data.seed_db.filter_seeds_by_tag(tags=flags)
        self.table_data.add_seed(data)


class SeedTableFilterBox(QLineEdit):
    def __init__(self, parent):
        super().__init__()

        self._table = parent

        self.installEventFilter(self)

    def eventFilter(self, obj, event):  # pylint:disable=unused-argument
        if event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_Escape:
                if self.text():
                    # clear the text
                    self.setText("")
                return True

        return False

class SeedTablePlugin(BasePlugin):
    """
    Plugin loader
    """

    def __init__(self, workspace):
        super().__init__(workspace)
        self.seed_table_view = SeedTableView(workspace, "center")
        workspace.default_tabs += [self.seed_table_view]
        workspace.add_view(self.seed_table_view)




# pylint:disable=unused-argument
import typing
from typing import List, Optional, TYPE_CHECKING
import threading

import asyncio
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

import PySide6
from PySide6.QtWidgets import QDialog, QPushButton, QHBoxLayout, QVBoxLayout, QMessageBox, QTableView, \
    QAbstractItemView, QHeaderView, QLabel
from PySide6.QtCore import Qt, QAbstractTableModel

try:
    import slacrs
except ImportError:
    slacrs = None

from angrmanagement.logic.threads import gui_thread_schedule_async

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class ChessTarget:
    """
    Models a CHESS challenge target.
    """
    def __init__(self, description: str, target_id: str, challenge_name: str, image_id: str):
        self.description = description
        self.target_id = target_id
        self.challenge_name = challenge_name
        self.image_id = image_id


class QTargetSelectorTableModel(QAbstractTableModel):
    """
    Implements a table model for targets.
    """

    Headers = ["Description", "Challenge", "Image ID"]
    COL_DESCRIPTION = 0
    COL_CHALLENGE = 1
    COL_IMAGEID = 2

    def __init__(self):
        super().__init__()
        self._targets: List[ChessTarget] = [ ]

    @property
    def targets(self):
        return self._targets

    @targets.setter
    def targets(self, v):
        self.beginResetModel()
        self._targets = v
        self.endResetModel()

    def rowCount(self, parent:PySide6.QtCore.QModelIndex=...) -> int:
        return len(self.targets)

    def columnCount(self, parent:PySide6.QtCore.QModelIndex=...) -> int:
        return len(self.Headers)

    def headerData(self, section:int, orientation:PySide6.QtCore.Qt.Orientation, role:int=...) -> typing.Any:
        if role != Qt.DisplayRole:
            return None

        if section < len(self.Headers):
            return self.Headers[section]
        return None

    def data(self, index:PySide6.QtCore.QModelIndex, role:int=...) -> typing.Any:
        if not index.isValid():
            return None
        row = index.row()
        if row >= len(self.targets):
            return None
        target = self.targets[row]
        col = index.column()

        if role == Qt.DisplayRole:
            return self._get_column_text(target, col)

        return None

    @staticmethod
    def _get_column_text(target: ChessTarget, col: int) -> str:
        mapping = {
            QTargetSelectorTableModel.COL_DESCRIPTION: QTargetSelectorTableModel._get_description,
            QTargetSelectorTableModel.COL_CHALLENGE: QTargetSelectorTableModel._get_challenge_name,
            QTargetSelectorTableModel.COL_IMAGEID: QTargetSelectorTableModel._get_image_id,
        }
        return mapping[col](target)

    @staticmethod
    def _get_description(target: ChessTarget) -> str:
        return target.description

    @staticmethod
    def _get_challenge_name(target: ChessTarget) -> str:
        return target.challenge_name

    @staticmethod
    def _get_image_id(target: ChessTarget) -> str:
        return target.image_id


class QTargetSelectorTableView(QTableView):
    """
    Implements a table view for targets.
    """
    def __init__(self):
        super().__init__()

        self.horizontalHeader().setVisible(True)
        self.verticalHeader().setVisible(False)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)

        self.model: QTargetSelectorTableModel = QTargetSelectorTableModel()
        self.setModel(self.model)


class QTargetSelectorDialog(QDialog):
    """
    Implements a CHESS target selector dialog.
    """
    def __init__(self, workspace: 'Workspace', parent=None):
        super().__init__(parent)

        if slacrs is None:
            QMessageBox.Critical(self,
                                 "Slacrs is not installed",
                                 "Cannot import slacrs. Please make sure slacrs is properly installed.",
                                 QMessageBox.Ok)
            self.close()
            return

        self.workspace = workspace
        self.target_id: Optional[str] = None
        self.target_image_id: Optional[str] = None
        self.target_description: Optional[str] = None
        self.ok: bool = False
        self.setMinimumWidth(400)

        self._status_label: QLabel = None
        self._table: QTargetSelectorTableView = None
        self._ok_button: QPushButton = None
        self._cancel_button: QPushButton = None

        self._init_widgets()

        self._status_label.setText("Loading...")
        self.workspace.main_window.app.processEvents()
        th = threading.Thread(target=self._load_targets, daemon=True)
        th.start()

    def _init_widgets(self):

        # table
        self._table = QTargetSelectorTableView()

        # status
        status_lbl = QLabel("Status:")
        self._status_label = QLabel()
        status_layout = QHBoxLayout()
        status_layout.addWidget(status_lbl)
        status_layout.addWidget(self._status_label)
        status_layout.addStretch(0)

        # buttons
        self._ok_button = QPushButton("Ok")
        self._ok_button.clicked.connect(self._on_ok_button_clicked)
        self._cancel_button = QPushButton("Cancel")
        self._cancel_button.clicked.connect(self._on_cancel_button_clicked)
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self._ok_button)
        buttons_layout.addWidget(self._cancel_button)

        layout = QVBoxLayout()
        layout.addWidget(self._table)
        layout.addLayout(status_layout)
        layout.addLayout(buttons_layout)
        self.setLayout(layout)

    def _load_targets(self):
        from slacrs.model import Target, Challenge  # pylint:disable=import-outside-toplevel,import-error
        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())

        connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
        if connector is None:
            # chess connector does not exist
            return
        slacrs_instance = connector.slacrs_instance()
        if slacrs_instance is None:
            # slacrs does not exist. continue
            return
        session = slacrs_instance.session()
        db_targets = session.query(Target)
        targets: List[ChessTarget] = [ ]

        for db_target in db_targets:
            db_challenge = session.query(Challenge).filter(
                Challenge.id == db_target.challenge_id
            ).first()
            t = ChessTarget(db_target.description, db_target.id, db_challenge.name, db_target.images[0].id)
            targets.append(t)

        session.close()
        gui_thread_schedule_async(self._update_table, args=(targets,))

    def _update_table(self, targets):
        self._table.model.targets = targets
        self._table.viewport().update()
        self._status_label.setText("Ready.")

    #
    # Events
    #

    def _on_ok_button_clicked(self):

        selection_model = self._table.selectionModel()
        if not selection_model.hasSelection():
            QMessageBox.warning(self,
                                "No target is selected",
                                "Please select a CHESS target to continue.",
                                QMessageBox.Ok)
            return

        rows = selection_model.selectedRows()
        target = self._table.model.targets[rows[0].row()]

        self.target_id = target.target_id
        self.target_description = target.description
        self.target_image_id = target.image_id
        self.ok = True
        self.close()

    def _on_cancel_button_clicked(self):
        self.target_id = None
        self.target_description = None
        self.ok = False
        self.close()



from typing import Optional, Dict, List, Tuple, TYPE_CHECKING
import os
import json
import binascii

from PySide6.QtWidgets import (
    QDialog,
    QLineEdit,
    QLabel,
    QHBoxLayout,
    QVBoxLayout,
    QPushButton,
    QProgressBar,
    QMessageBox,
    QSizePolicy,
    QFileDialog,
)

from ....utils.io import isurl, download_url
from ..data.component_tree import ComponentTree, ComponentTreeNode, ComponentFunction

if TYPE_CHECKING:
    from ....ui.workspace import Workspace


class LoadComponentsDialog(QDialog):
    def __init__(self, workspace: 'Workspace', url: Optional[str]=None, parent=None):
        super().__init__(parent)

        self.url_box: QLineEdit = None
        self.ok_btn: QPushButton = None
        self.progressbar: QProgressBar = None

        self.setWindowTitle("Load components...")

        self.workspace = workspace
        self.url = url
        self.tree: ComponentTree = None

        self._init_widgets()

        if self.url is not None:
            self.url_box.setText(self.url)

    #
    # Actions
    #

    def exec_(self) -> int:

        if self.url is not None:
            self.show()
            self._on_ok_clicked()
            if self.tree is not None:
                return 0

        return super().exec_()

    def load_json(self):
        path = self.url_box.text()
        try:
            if isurl(path):
                tree = self._load_json_from_url(path)
            else:
                tree = self._load_json_from_file(path)
            self.tree = tree
            return True
        except (ValueError, TypeError, IOError) as ex:
            QMessageBox.critical(self,
                                 "Loading error",
                                 "Failed to load components information. Please make sure the component file is "
                                 "valid.\nException: %s" % ex,
                                 QMessageBox.Ok)
        return False

    def close_dialog(self):
        self.close()

    #
    # Initialization
    #

    def _init_widgets(self):

        # URL
        url_caption = QLabel(self)
        url_caption.setText("Path or URL:")

        self.url_box = QLineEdit(self)
        self.url_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        btn = QPushButton(self)
        btn.setText("...")
        btn.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        btn.clicked.connect(self._on_browse_file_btn_clicked)

        url_layout = QHBoxLayout()
        url_layout.addWidget(url_caption)
        url_layout.addWidget(self.url_box)
        url_layout.addWidget(btn)

        # progress bar
        self.progressbar = QProgressBar()
        self.progressbar.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        # buttons
        self.ok_btn = QPushButton(self)
        self.ok_btn.setText("&Load components")
        self.ok_btn.clicked.connect(self._on_ok_clicked)

        close_btn = QPushButton(self)
        close_btn.setText("&Close")
        close_btn.clicked.connect(self._on_close_clicked)

        btn_layout = QHBoxLayout()
        btn_layout.addWidget(self.ok_btn)
        btn_layout.addWidget(close_btn)

        # overall layout
        layout = QVBoxLayout()
        layout.addLayout(url_layout)
        layout.addWidget(self.progressbar)
        layout.addLayout(btn_layout)

        self.setLayout(layout)

    #
    # JSON loading
    #

    def _load_json_from_file(self, path):
        if not os.path.isfile(path):
            raise IOError("File %s does not exist." % path)

        with open(path, "r") as f:
            try:
                data = json.load(f)
            except ValueError as ex:
                raise TypeError("File %s does not contain valid JSON data.\nException: %s." % (path, ex))

        return self._load_json(data)

    def _load_json_from_url(self, url):
        content = download_url(url, parent=self, to_file=False)

        try:
            data = json.loads(content.decode("utf-8"))
        except ValueError as ex:
            raise TypeError("URL %s does not contain valid JSON data.\nException: %s." % (url, ex))
        return self._load_json(data)

    def _load_json(self, data: List[Dict]):

        # ok let's do this
        tree = ComponentTree()
        queue: List[Tuple[Dict, Optional[ComponentTreeNode]]] = [ (child, None) for child in data ]

        for node, parent in queue:
            label = node['label']
            if not label:
                # root?
                if parent is not None:
                    raise TypeError("Found a root node with a parent node.")

                binary = node['binary']
                blob_type = binary.get('blob_type')
                if blob_type != "bin":
                    raise TypeError("Unsupported blob_type \"%s\"." % blob_type)
                file_hash = binary.get('file_hash')
                if self.workspace.main_instance.project.am_none:
                    raise ValueError("No project has been loaded.")
                sha256 = self.workspace.main_instance.project.loader.main_object.sha256
                if binascii.unhexlify(file_hash) != sha256:
                    # warn user about it
                    r = QMessageBox.warning(None,
                                            "File hash mismatch",
                                            "The SHA256 hash of the main object is:\n"
                                            "    {main_obj_hash},\n"
                                            "while the ShA256 hash of the given components file is:\n"
                                            "    {hash}.\n"
                                            "Do you want to continue loading the components from the "
                                            "given JSON file?".format(
                                                main_obj_hash=binascii.hexlify(sha256),
                                                hash=file_hash,
                                            ),
                                            QMessageBox.Yes | QMessageBox.No)
                    if r == QMessageBox.No:
                        return False

                root = ComponentTreeNode(name="Root")
                tree.root = root

                for child in node['children']:
                    queue.append((child, root))

            else:
                tree_node = ComponentTreeNode(label)
                tree_node.functions = [ self._load_function(f) for f in node['functions'] ]
                for child in node['children']:
                    queue.append((child, tree_node))
                if parent is None:
                    if tree.root is not None:
                        raise ValueError("Found more than one root in the JSON file.")
                    tree.root = tree_node
                else:
                    parent.components.append(tree_node)

        return tree

    def _load_function(self, func: Dict):
        rebased_addr = func['start']['rebased_addr']
        f_node = ComponentFunction(rebased_addr['mapped_base'],
                                   rebased_addr['virtual_addr'],
                                   symbol_name=func.get('symbol_name', None))
        return f_node

    #
    # Events
    #

    def _on_ok_clicked(self):
        if self.load_json():
            self.close_dialog()

    def _on_close_clicked(self):
        self.close_dialog()

    def _on_browse_file_btn_clicked(self):
        file_path, _ = QFileDialog.getOpenFileName(self, "Open a components JSON file", "",
                                                   "All files (*);;JSON file (*.json)",
                                                   )
        if file_path is not None:
            self.url_box.setText(file_path)



import logging
import subprocess

from PySide6.QtWidgets import QInputDialog, QMessageBox


_l = logging.getLogger(__name__)


class LoadDockerPromptError(Exception):
    pass


class LoadDockerPrompt(QInputDialog):
    def __init__(self, parent=None):
        super(LoadDockerPrompt, self).__init__(parent)

        self.setComboBoxItems(get_docker_images(self))
        self.setOption(QInputDialog.UsePlainTextEditForTextInput, True)
        self.setLabelText('Select a docker target')
        self.setComboBoxEditable(False)
        self.setModal(True)
        _l.info('Docker prompt succesfully loaded')

def get_docker_images(parent=None):
    # Try to list docker images, handle failure cases
    try:
        output = subprocess.check_output(['docker', 'images', '--format', '{{.Repository}}:{{.Tag}}'])
    except FileNotFoundError as e:
        _l.error('Docker not found.')
        QMessageBox(parent).critical(
            None,
            'Docker not found',
            'Make sure docker is installed and in your path'
        )
        raise LoadDockerPromptError from e
    except subprocess.CalledProcessError as e:
        _l.error('Docker images failed: Make sure you are have privileges.')
        QMessageBox(parent).critical(
            None,
            'Docker Error',
            'User not able to access docker. Are you in the docker group?'
        )
        raise LoadDockerPromptError from e

    items = output.decode('utf-8').split('\n')
    return [ i for i in items if i and '<none>' not in i ]



# pylint:disable=import-outside-toplevel,unused-import,no-member
import asyncio
import sys
import os
import ctypes
import threading
import time
import warnings
import platform
import signal

from . import __version__

if sys.platform.startswith("darwin"):
    from Foundation import NSBundle  # pylint: disable=import-error


def shut_up(*args, **kwargs):  # pylint:disable=unused-argument
    return
warnings.simplefilter = shut_up


BUGGY_PYSIDE2_VERSIONS = [
    "5.12.1",  # https://github.com/angr/angr-management/issues/59
    "5.14.2",  # Tiffany reported. Verified. Probably caused by
               # https://code.qt.io/cgit/pyside/pyside-setup.git/commit/?h=5.14&id=52299827c64cccc1456f9050fdf3dd8596df3e6f
    "5.14.2.1",  # deadlocks sometimes, although better than 5.14.2
]


name: str = "angr management"


def check_dependencies_qt():

    missing_dep = False

    try:
        import PySide6
    except ImportError:
        PySide6 = None
        sys.stderr.write("Cannot find the PySide6 package. You may install it via pip:\n" +
                         "    pip install pyside6\n")
        missing_dep = True

    try:
        import qtconsole
    except ImportError:
        sys.stderr.write("Cannot find the qtconsole package. You may install it via pip:\n" +
                         "    pip install qtconsole\n")
        missing_dep = True

    return not missing_dep


def check_dependencies():

    missing_dep = False

    try:
        import sqlalchemy
    except ImportError:
        sys.stderr.write("Cannot find the sqlalchemy package. You may install it via pip:\n" +
                         "    pip install sqlalchemy\n")
        missing_dep = True

    try:
        import pyqodeng.core
    except ImportError:
        sys.stderr.write("Cannot find the pyqodeng.core package. You may install it via pip:\n" +
                         "    pip install pyqodeng.core\n")
        missing_dep = True

    try:
        import xdg
    except ImportError:
        sys.stderr.write("Cannot find the xdg package. You may install it via pip:\n" +
                         "    pip install pyxdg\n")
        missing_dep = True

    return not missing_dep


def set_app_user_model_id():
    # Explicitly call SetCurrentProcessExplicitAppUserModelID() so the taskbar icon is displayed correctly.

    if sys.platform == 'win32':
        winver = sys.getwindowsversion()
        if winver.major >= 5:
            myappid = name.replace(" ", "-")
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)


def set_windows_event_loop_policy():
    if sys.platform == 'win32':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def macos_bigsur_wants_layer():
    # workaround for https://bugreports.qt.io/browse/QTBUG-87014
    # this is because the latest PySide6 (5.15.2) does not include this fix
    v, _, _ = platform.mac_ver()
    vs = v.split(".")
    if len(vs) >= 2:
        major, minor = [int(x) for x in vs[:2]]
    else:
        return
    if major >= 11 or major == 10 and minor == 16:
        os.environ['QT_MAC_WANTS_LAYER'] = '1'


def start_management(filepath=None, use_daemon=None, profiling=False):

    if sys.platform == "darwin":
        macos_bigsur_wants_layer()

    if not check_dependencies_qt():
        # it's likely that other dependencies are also missing. check them here before exiting.
        check_dependencies()
        sys.exit(1)

    set_app_user_model_id()
    set_windows_event_loop_policy()

    from PySide6.QtWidgets import QApplication, QSplashScreen
    from PySide6.QtGui import QFontDatabase, QPixmap, QIcon
    from PySide6.QtCore import Qt, QCoreApplication

    from .config import FONT_LOCATION, IMG_LOCATION, Conf

    # Enable High-DPI support
    # https://stackoverflow.com/questions/35714837/how-to-get-sharp-ui-on-high-dpi-with-qt-5-6
    if ("QT_DEVICE_PIXEL_RATIO" not in os.environ
        and "QT_AUTO_SCREEN_SCALE_FACTOR" not in os.environ
        and "QT_SCALE_FACTOR" not in os.environ
        and "QT_SCREEN_SCALE_FACTORS" not in os.environ
        ):
        QCoreApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    # No more rounding
    # https://github.com/pyqtgraph/pyqtgraph/issues/756
    # https://lists.qt-project.org/pipermail/development/2019-September/037434.html
    QApplication.setHighDpiScaleFactorRoundingPolicy(Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)
    # Use highDPI pixmaps
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)

    # Fix app title on macOS
    if sys.platform.startswith("darwin"):
        try:
            bundle = NSBundle.mainBundle()
            info = bundle.localizedInfoDictionary() or bundle.infoDictionary()
            info["CFBundleName"] = name
        except Exception as e:  # pylint: disable=broad-except
            # This happens before logging is setup so use stderr
            print(f"{type(e).__name__}: {e}", file=sys.stderr)

    app = QApplication(sys.argv)
    app.setApplicationDisplayName(name)
    app.setApplicationName(name)
    icon_location = os.path.join(IMG_LOCATION, 'angr.png')
    QApplication.setWindowIcon(QIcon(icon_location))

    # try to import the initial configuration for the install
    Conf.attempt_importing_initial_config()

    # Make + display splash screen
    splashscreen_location = os.path.join(IMG_LOCATION, 'angr-splash.png')
    splash_pixmap = QPixmap(splashscreen_location)
    splash = QSplashScreen(splash_pixmap, Qt.WindowStaysOnTopHint)
    icon_location = os.path.join(IMG_LOCATION, 'angr.png')
    splash.setWindowIcon(QIcon(icon_location))
    splash.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.FramelessWindowHint)
    splash.setEnabled(False)
    splash.show()
    for _ in range(5):
        time.sleep(0.01)
        app.processEvents()

    if not check_dependencies():
        sys.exit(1)

    from .ui.css import refresh_theme  # import .ui after showing the splash screen since it's going to take time

    refresh_theme()

    import angr

    angr.loggers.profiling_enabled = bool(profiling)

    from .logic import GlobalInfo
    from .ui.main_window import MainWindow

    # Load fonts
    QFontDatabase.addApplicationFont(os.path.join(FONT_LOCATION, "SourceCodePro-Regular.ttf"))
    QFontDatabase.addApplicationFont(os.path.join(FONT_LOCATION, "DejaVuSansMono.ttf"))

    # Initialize font-related configuration
    Conf.init_font_config()
    # Set global font
    app.setFont(Conf.ui_default_font)

    GlobalInfo.gui_thread = threading.get_ident()

    file_to_open = filepath if filepath else None
    main_window = MainWindow(app=app, use_daemon=use_daemon)
    splash.finish(main_window)

    if file_to_open is not None:
        main_window.load_file(file_to_open)

    main_window.initialized = True
    main_window.workspace.view_manager.main_window_initialized()

    app.exec_()


def main():
    import argparse
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    prog: str = name.replace(" ", "-")
    parser = argparse.ArgumentParser(prog=prog, description=name)
    parser.add_argument("-v", "--version", action="version", version=f"{prog} {__version__}")
    parser.add_argument("-s", "--script", type=str, help="run a python script in the (commandline) angr environment")
    parser.add_argument("-i", "--interactive", action='store_true', help="interactive (ipython) mode")
    parser.add_argument("-n", "--no-gui", action='store_true', help="run in headless mode")
    parser.add_argument('-d', "--with-daemon", action='store_true', help="use angr with the daemon. this allows angr "
                                                                         "to handle angr:// URLs. it will "
                                                                         "automatically start a daemon if there isn't "
                                                                         "already one running.")
    parser.add_argument("-D", "--daemon", action='store_true', help="start a daemon to handle angr:// URLs.")
    parser.add_argument("-u", "--url", type=str, nargs='?', help="(internal) handle angr:// URLs. "
                                                                 "the daemon must be running.")
    parser.add_argument("-p", "--profiling", action='store_true', help="display profiling log messages.")
    parser.add_argument('-R', '--autoreload', action='store_true', help="Reload all python modules on each job start.")
    parser.add_argument("binary", nargs="?", help="the binary to open (for the GUI)")

    args = parser.parse_args()

    if args.autoreload:
        from .logic import GlobalInfo
        GlobalInfo.autoreload = True

    if args.daemon:
        from .daemon import start_daemon
        start_daemon()
        return
    elif args.url:
        from .daemon import daemon_exists, handle_url, run_daemon_process, daemon_conn
        if not daemon_exists():
            run_daemon_process()
            time.sleep(1)

        # initialize plugins
        from .plugins import PluginManager
        PluginManager(None).discover_and_initialize_plugins()

        action = handle_url(args.url, act=False)
        action.act(daemon_conn())
        return
    if args.script:
        import runpy
        script_globals = runpy.run_path(args.script)  # pylint:disable=unused-variable
    if args.interactive:
        if args.script:
            print("Your script's globals() dict is available in the `script_globals` variable.")
        import IPython
        IPython.embed(banner1="")
    if not args.no_gui:
        start_management(args.binary,
                         use_daemon=True if args.with_daemon else None,
                         profiling=True if args.profiling else None)


if __name__ == '__main__':
    main()




import os

from PySide6.QtGui import QIcon

from ...config import IMG_LOCATION
from .toolbar import Toolbar, ToolbarAction


class FunctionTableToolbar(Toolbar):
    def __init__(self, function_table):
        super().__init__(function_table, 'Function table options')

        # TODO: An icon would be great
        self._alignment_action = ToolbarAction(QIcon(os.path.join(IMG_LOCATION, 'toolbar-show-alignment.png')),
                                               "Show alignment functions",
                                               "Display alignment function stubs.",
                                               function_table.toggle_show_alignment_functions,
                                               checkable=True)

        self.actions = [
            self._alignment_action,
        ]

    def toggle_show_alignment_functions(self):
        self.window.toggle_show_alignment_functions()

        if self._cached_actions and self._alignment_action in self._cached_actions:
            self._cached_actions[self._alignment_action].setChecked(self.window.show_alignment_functions)
